@@ -23,6 +23,8 @@ import inspect
 import os
 import sys
 
+from robot.utils import PY_VERSION
+
 try:
     import typing
 except ImportError:
@@ -34,24 +36,27 @@ from robot import __version__ as robot_version
 PY2 = sys.version_info < (3,)
 RF31 = robot_version < '3.2'
 
-__version__ = '2.0.2'
+__version__ = '2.2.0'
 
 
 class HybridCore(object):
 
     def __init__(self, library_components):
         self.keywords = {}
+        self.keywords_spec = {}
         self.attributes = {}
         self.add_library_components(library_components)
         self.add_library_components([self])
 
     def add_library_components(self, library_components):
+        self.keywords_spec['__init__'] = KeywordBuilder.build(self.__init__)
         for component in library_components:
             for name, func in self.__get_members(component):
                 if callable(func) and hasattr(func, 'robot_name'):
                     kw = getattr(component, name)
                     kw_name = func.robot_name or name
                     self.keywords[kw_name] = kw
+                    self.keywords_spec[kw_name] = KeywordBuilder.build(kw)
                     # Expose keywords as attributes both using original
                     # method names as well as possible custom names.
                     self.attributes[name] = self.attributes[kw_name] = kw
@@ -98,11 +103,8 @@ class DynamicCore(HybridCore):
         return self.keywords[name](*args, **(kwargs or {}))
 
     def get_keyword_arguments(self, name):
-        kw_method = self.__get_keyword(name)
-        if kw_method is None:
-            return None
-        spec = ArgumentSpec.from_function(kw_method)
-        return spec.get_arguments()
+        spec = self.keywords_spec.get(name)
+        return spec.argument_specification
 
     def get_keyword_tags(self, name):
         return self.keywords[name].robot_tags
@@ -110,22 +112,14 @@ class DynamicCore(HybridCore):
     def get_keyword_documentation(self, name):
         if name == '__intro__':
             return inspect.getdoc(self) or ''
-        if name == '__init__':
-            return inspect.getdoc(self.__init__) or ''
-        kw = self.keywords[name]
-        return inspect.getdoc(kw) or ''
+        spec = self.keywords_spec.get(name)
+        return spec.documentation
 
-    def get_keyword_types(self, keyword_name):
-        method = self.__get_keyword(keyword_name)
-        if method is None:
-            return method
-        types = getattr(method, 'robot_types', ())
-        if types is None:
-            return types
-        if not types:
-            types = self.__get_typing_hints(method)
-        types = self.__join_defaults_with_types(method, types)
-        return types
+    def get_keyword_types(self, name):
+        spec = self.keywords_spec.get(name)
+        if spec is None:
+            raise ValueError('Keyword "%s" not found.' % name)
+        return spec.argument_types
 
     def __get_keyword(self, keyword_name):
         if keyword_name == '__init__':
@@ -137,22 +131,6 @@ class DynamicCore(HybridCore):
             raise ValueError('Keyword "%s" not found.' % keyword_name)
         return method
 
-    def __get_typing_hints(self, method):
-        if PY2:
-            return {}
-        spec = ArgumentSpec.from_function(method)
-        return spec.get_typing_hints()
-
-    def __join_defaults_with_types(self, method, types):
-        spec = ArgumentSpec.from_function(method)
-        for name, value in spec.defaults:
-            if name not in types and isinstance(value, (bool, type(None))):
-                types[name] = type(value)
-        for name, value in spec.kwonlydefaults:
-            if name not in types and isinstance(value, (bool, type(None))):
-                types[name] = type(value)
-        return types
-
     def get_keyword_source(self, keyword_name):
         method = self.__get_keyword(keyword_name)
         path = self.__get_keyword_path(method)
@@ -182,97 +160,157 @@ class DynamicCore(HybridCore):
             return None
 
 
-class StaticCore(HybridCore):
+class KeywordBuilder(object):
 
-    def __init__(self):
-        HybridCore.__init__(self, [])
+    @classmethod
+    def build(cls, function):
+        if not PY2:
+            function = inspect.unwrap(function)
+        return KeywordSpecification(
+            argument_specification=cls._get_arguments(function),
+            documentation=inspect.getdoc(function) or '',
+            argument_types=cls._get_types(function)
+        )
 
+    @classmethod
+    def _get_arguments(cls, function):
+        arg_spec = cls._get_arg_spec(function)
+        argument_specification = cls._get_default_and_named_args(
+            arg_spec, function
+        )
+        argument_specification.extend(cls._get_var_args(arg_spec))
+        kw_only_args = cls._get_kw_only(arg_spec)
+        if kw_only_args:
+            argument_specification.extend(kw_only_args)
+        argument_specification.extend(cls._get_kwargs(arg_spec))
+        return argument_specification
 
-class ArgumentSpec(object):
+    @classmethod
+    def _get_arg_spec(cls, function):
+        if PY2:
+            return inspect.getargspec(function)
+        return inspect.getfullargspec(function)
 
-    _function = None
+    @classmethod
+    def _get_default_and_named_args(cls, arg_spec, function):
+        args = cls._drop_self_from_args(function, arg_spec)
+        args.reverse()
+        defaults = list(arg_spec.defaults) if arg_spec.defaults else []
+        formated_args = []
+        for arg in args:
+            if defaults:
+                formated_args.append(
+                    cls._format_defaults(arg, defaults.pop())
+                )
+            else:
+                formated_args.append(arg)
+        formated_args.reverse()
+        return formated_args
 
-    def __init__(self, positional=None, defaults=None, varargs=None, kwonlyargs=None,
-                 kwonlydefaults=None, kwargs=None):
-        self.positional = positional or []
-        self.defaults = defaults or []
-        self.varargs = varargs
-        self.kwonlyargs = kwonlyargs or []
-        self.kwonlydefaults = kwonlydefaults or []
-        self.kwargs = kwargs
+    @classmethod
+    def _drop_self_from_args(cls, function, arg_spec):
+        return arg_spec.args[1:] if inspect.ismethod(function) else arg_spec.args
 
-    def __contains__(self, item):
-        if item in self.positional:
-            return True
-        if self.varargs and item in self.varargs:
-            return True
-        if item in self.kwonlyargs:
-            return True
-        if self.kwargs and item in self.kwargs:
-            return True
-        return False
+    @classmethod
+    def _get_var_args(cls, arg_spec):
+        if arg_spec.varargs:
+            return ['*%s' % arg_spec.varargs]
+        return []
 
-    def get_arguments(self):
-        args = self._format_positional(self.positional, self.defaults)
-        args += self._format_default(self.defaults)
-        if self.varargs:
-            args.append('*%s' % self.varargs)
-        args += self._format_positional(self.kwonlyargs, self.kwonlydefaults)
-        args += self._format_default(self.kwonlydefaults)
-        if self.kwargs:
-            args.append('**%s' % self.kwargs)
-        return args
+    @classmethod
+    def _get_kwargs(cls, arg_spec):
+        if PY2:
+            return ['**%s' % arg_spec.keywords] if arg_spec.keywords else []
+        return ['**%s' % arg_spec.varkw] if arg_spec.varkw else []
 
-    def get_typing_hints(self):
+    @classmethod
+    def _get_kw_only(cls, arg_spec):
+        kw_only_args = []
+        if PY2:
+            return kw_only_args
+        for arg in arg_spec.kwonlyargs:
+            if not arg_spec.kwonlydefaults or arg not in arg_spec.kwonlydefaults:
+                kw_only_args.append(arg)
+            else:
+                value = arg_spec.kwonlydefaults.get(arg, '')
+                kw_only_args.append(cls._format_defaults(arg, value))
+        return kw_only_args
+
+    @classmethod
+    def _format_defaults(cls, arg, value):
+        if RF31:
+            return '%s=%s' % (arg, value)
+        return arg, value
+
+    @classmethod
+    def _get_types(cls, function):
+        if function is None:
+            return function
+        types = getattr(function, 'robot_types', ())
+        if types is None or types:
+            return types
+        return cls._get_typing_hints(function)
+
+    @classmethod
+    def _get_typing_hints(cls, function):
+        if PY2:
+            return {}
         try:
-            hints = typing.get_type_hints(self._function)
+            hints = typing.get_type_hints(function)
         except Exception:
-            hints = self._function.__annotations__
-        for arg in list(hints):
+            hints = function.__annotations__
+        arg_spec = cls._get_arg_spec(function)
+        all_args = cls._args_as_list(function, arg_spec)
+        for arg_with_hint in list(hints):
             # remove return and self statements
-            if arg not in self:
-                hints.pop(arg)
-        return hints
-
-    def _format_positional(self, positional, defaults):
-        for argument, _ in defaults:
-            positional.remove(argument)
-        return positional
-
-    def _format_default(self, defaults):
-        if not RF31:
-            return [default for default in defaults]
-        return ['%s=%s' % (argument, default) for argument, default in defaults]
+            if arg_with_hint not in all_args:
+                hints.pop(arg_with_hint)
+        default = cls._get_defaults(arg_spec)
+        return cls._remove_optional_none_type_hints(hints, default)
 
     @classmethod
-    def from_function(cls, function):
-        cls._function = function
-        if PY2:
-            spec = inspect.getargspec(function)
-        else:
-            spec = inspect.getfullargspec(function)
-        args = spec.args[1:] if inspect.ismethod(function) else spec.args  # drop self
-        defaults = cls._get_defaults(spec)
-        kwonlyargs, kwonlydefaults, kwargs = cls._get_kw_args(spec)
-        return cls(positional=args,
-                   defaults=defaults,
-                   varargs=spec.varargs,
-                   kwonlyargs=kwonlyargs,
-                   kwonlydefaults=kwonlydefaults,
-                   kwargs=kwargs)
+    def _args_as_list(cls, function, arg_spec):
+        function_args = []
+        function_args.extend(cls._drop_self_from_args(function, arg_spec))
+        if arg_spec.varargs:
+            function_args.append(arg_spec.varargs)
+        function_args.extend(arg_spec.kwonlyargs or [])
+        if arg_spec.varkw:
+            function_args.append(arg_spec.varkw)
+        return function_args
+
+    # Copied from: robot.running.arguments.argumentparser
+    @classmethod
+    def _remove_optional_none_type_hints(cls, type_hints, defaults):
+        # If argument has None as a default, typing.get_type_hints adds
+        # optional None to the information it returns. We don't want that.
+        for arg, default in defaults:
+            if default is None and arg in type_hints:
+                type_ = type_hints[arg]
+                if cls._is_union(type_):
+                    types = type_.__args__
+                    if len(types) == 2 and types[1] is type(None): # noqa
+                        type_hints[arg] = types[0]
+        return type_hints
+
+    # Copied from: robot.running.arguments.argumentparser
+    @classmethod
+    def _is_union(cls, typing_type):
+        if PY_VERSION >= (3, 7) and hasattr(typing_type, '__origin__'):
+            typing_type = typing_type.__origin__
+        return isinstance(typing_type, type(typing.Union))
 
     @classmethod
-    def _get_defaults(cls, spec):
-        if not spec.defaults:
-            return []
-        names = spec.args[-len(spec.defaults):]
-        return list(zip(names, spec.defaults))
+    def _get_defaults(cls, arg_spec):
+        if not arg_spec.defaults:
+            return {}
+        names = arg_spec.args[-len(arg_spec.defaults):]
+        return zip(names, arg_spec.defaults)
 
-    @classmethod
-    def _get_kw_args(cls, spec):
-        if PY2:
-            return [], [], spec.keywords
-        kwonlyargs = spec.kwonlyargs or []
-        defaults = spec.kwonlydefaults or {}
-        kwonlydefaults = [(arg, name) for arg, name in defaults.items()]
-        return kwonlyargs, kwonlydefaults, spec.varkw
+
+class KeywordSpecification(object):
+
+    def __init__(self, argument_specification=None, documentation=None, argument_types=None):
+        self.argument_specification = argument_specification
+        self.documentation = documentation
+        self.argument_types = argument_types
